"""Bindings for different external packages.

3 Bindings are available:

* A `Nipype`<https://github.com/nipype/nipype/>_ binding
* A `Modopt`<https://cea-cosmic.github.io/ModOpt/index.html>_ binding
* A Command line interface, available as ``patch-denoise`` after installation.
"""
